"""A module controlling the app."""
import logging
from dataclasses import fields, is_dataclass
from typing import Any, Callable, Iterable, List, Optional, Type

from click import confirm
from epplib.commands.base import Hello, PollAcknowledgement, PollRequest, RawRequest, Request
from epplib.commands.check import CheckContact, CheckDomain, CheckKeyset, CheckNsset
from epplib.commands.command_extensions import (CreateContactMailingAddressExtension, CreateDomainEnumExtension,
                                                RenewDomainEnumExtension, UpdateContactMailingAddressExtension,
                                                UpdateDomainEnumExtension)
from epplib.commands.create import CreateContact, CreateDomain, CreateKeyset, CreateNsset
from epplib.commands.delete import DeleteContact, DeleteDomain, DeleteKeyset, DeleteNsset
from epplib.commands.extensions import (CreditInfoRequest, SendAuthInfoContact, SendAuthInfoDomain, SendAuthInfoKeyset,
                                        SendAuthInfoNsset, TestNsset)
from epplib.commands.info import InfoContact, InfoDomain, InfoKeyset, InfoNsset
from epplib.commands.list import (GetResults, ListContacts, ListDomains, ListDomainsByContact, ListDomainsByKeyset,
                                  ListDomainsByNsset, ListKeysets, ListKeysetsByContact, ListNssets,
                                  ListNssetsByContact, ListNssetsByNs)
from epplib.commands.renew import RenewDomain
from epplib.commands.transfer import TransferContact, TransferDomain, TransferKeyset, TransferNsset
from epplib.commands.update import UpdateContact, UpdateDomain, UpdateKeyset, UpdateNsset

from .commands.argument_registry import ArgumentRegistry, register_defaults
from .commands.command import Command
from .commands.command_registry import CommandRegistry
from .commands.help import HelpCommand
from .commands.hooks import apply_disclose_policy, get_results_hook
from .commands.service import (ServiceChangeSettings, ServiceCommand, ServiceExit, ServiceLogin, ServiceLogout,
                               ServiceNewSession)
from .epp import EppHandler
from .models import ArgumentField, ExecutableCommand
from .settings import Session, Settings

LOGGER = logging.getLogger(__name__)

CommandHook = Callable[['Application', ExecutableCommand], None]


class Application:
    """Main application information."""

    def __init__(self, settings: Settings, command: Optional[str],
                 session: Optional[Session] = None) -> None:
        """Load up settings."""
        self.argument_registry = ArgumentRegistry()
        register_defaults(self.argument_registry)
        self.command_registry = CommandRegistry()
        self.register_commands()

        self.input_handler = settings.input_format.get_input_handler(
            command, settings.input_file, self.command_registry,
            settings.interactive_input.history_file_path, self._prompt,
            settings.syntax_highlighting, settings.interactive_input.autocomplete)
        self.input_processor = settings.input_format.get_input_processor()

        self.epp_handler = EppHandler(session)

        self.set_settings(settings)

    def set_settings(self, settings: Settings) -> None:
        """Set `settings` and update related changable attributes."""
        self.settings = settings
        self.input_handler.set_input_preferences(settings.interactive_input.autocomplete,
                                                 settings.syntax_highlighting)
        self.output_handler = settings.output_format.get_output_handler(settings.syntax_highlighting)

    def add_session(self, session: Session) -> int:
        """Add a session to the app settings.

        This doesn't change the session set in the `epp_handler`.
        """
        self.settings.sessions.append(session)
        return len(self.settings.sessions) - 1

    def _prompt(self) -> str:
        """Get prompt: username@host."""
        if not self.epp_handler.logged_in:
            return "> "
        return f"{self.epp_handler.session.username}@{self.epp_handler.session.hostname} > "

    def register_command_to_registry(
            self, name: str, command_class: Type[ExecutableCommand],
            extensions: Optional[Iterable[ArgumentField]] = None,
            pre_hooks: Optional[List[CommandHook]] = None,
            post_hooks: Optional[List[CommandHook]] = None,
            number_of_positional_args: int = 0) -> None:
        """Create and register a command."""
        a_fields = (ArgumentField.from_dataclass_field(field) for field in fields(command_class)
                    ) if is_dataclass(command_class) else ()
        command = Command(name, command_class, a_fields, self.argument_registry, extensions,
                          pre_hooks, post_hooks, number_of_positional_args)
        self.command_registry.register_command(command)

    def execute(self, command: ExecutableCommand) -> None:
        """Execute the prepared command."""
        if isinstance(command, ServiceCommand):
            command.handle(self)
        else:
            assert isinstance(command, Request)  # nosec
            response = self.epp_handler.send_command(command)
            LOGGER.debug(response)
            self.output_handler.output(response)

    def confirm_command(self, command: ExecutableCommand) -> bool:
        """Confirm command."""
        if not isinstance(command, Request) or not self.settings.command_confirmation:
            return True

        print(command)
        print(command.xml())

        return confirm("Send command?")

    def run_command(self, inp: Any) -> None:
        """Run one command and stop."""
        LOGGER.debug("Parsing raw input...")
        cmd = self.input_processor(inp)
        LOGGER.debug("Parsing command...")
        command = self.command_registry.get_command(cmd.name)
        LOGGER.debug("Parsing arguments...")
        cmd = command.arg_parser.parse_arguments(cmd)
        if cmd.error is not None:
            raise ValueError(cmd.error)
        assert cmd.args is not None, "cmd.args is not None on processed command."  # nosec
        LOGGER.debug(f"Creating {cmd.name} command...")
        epp_command = command.prepare_command(cmd.args)
        LOGGER.debug("Verifying command...")
        if not self.confirm_command(epp_command):
            LOGGER.info("Command aborted.")
            return

        LOGGER.debug("Executing pre-hooks.")
        command.pre_hooks(self, epp_command)
        LOGGER.debug("Executing command...")
        self.execute(epp_command)
        LOGGER.debug("Executing post-hooks.")
        command.post_hooks(self, epp_command)
        LOGGER.debug("Command completed.")

    def run(self) -> None:
        """Run commands continuously."""
        for command in self.input_handler:
            try:
                self.run_command(command)
            except Exception as e:
                LOGGER.error(e)

        self.epp_handler.logout()

    def register_commands(self) -> None:
        """Return a list of all commands."""
        self.register_command_to_registry("hello", Hello)
        self.register_command_to_registry("exit", ServiceExit)
        self.register_command_to_registry("raw-request", RawRequest)

        # session
        self.register_command_to_registry("login", ServiceLogin,
                                          number_of_positional_args=2)
        self.register_command_to_registry("logout", ServiceLogout)
        self.register_command_to_registry("new-session", ServiceNewSession,
                                          number_of_positional_args=10)
        self.register_command_to_registry("change-settings", ServiceChangeSettings)

        self.register_command_to_registry("poll-req", PollRequest)
        self.register_command_to_registry("poll-ack", PollAcknowledgement,
                                          number_of_positional_args=1)

        # info
        self.register_command_to_registry("info-contact", InfoContact,
                                          number_of_positional_args=1)
        self.register_command_to_registry("info-domain", InfoDomain,
                                          number_of_positional_args=1)
        self.register_command_to_registry("info-keyset", InfoKeyset,
                                          number_of_positional_args=1)
        self.register_command_to_registry("info-nsset", InfoNsset,
                                          number_of_positional_args=1)

        # transfer
        self.register_command_to_registry("transfer-contact", TransferContact,
                                          number_of_positional_args=1)
        self.register_command_to_registry("transfer-domain", TransferDomain,
                                          number_of_positional_args=1)
        self.register_command_to_registry("transfer-keyset", TransferKeyset,
                                          number_of_positional_args=1)
        self.register_command_to_registry("transfer-nsset", TransferNsset,
                                          number_of_positional_args=1)

        # update
        self.register_command_to_registry("update-contact", UpdateContact,
                                          [ArgumentField("mailing", UpdateContactMailingAddressExtension, False)],
                                          pre_hooks=[apply_disclose_policy],
                                          number_of_positional_args=1)
        self.register_command_to_registry("update-domain", UpdateDomain,
                                          [ArgumentField("enum", UpdateDomainEnumExtension, False)],
                                          number_of_positional_args=1)
        self.register_command_to_registry("update-keyset", UpdateKeyset,
                                          number_of_positional_args=1)
        self.register_command_to_registry("update-nsset", UpdateNsset,
                                          number_of_positional_args=1)

        # create
        self.register_command_to_registry("create-contact", CreateContact,
                                          [ArgumentField("mailing", CreateContactMailingAddressExtension, False)],
                                          pre_hooks=[apply_disclose_policy],
                                          number_of_positional_args=1)
        self.register_command_to_registry("create-domain", CreateDomain,
                                          [ArgumentField("enum", CreateDomainEnumExtension, False)],
                                          number_of_positional_args=1)
        self.register_command_to_registry("create-keyset", CreateKeyset,
                                          number_of_positional_args=1)
        self.register_command_to_registry("create-nsset", CreateNsset,
                                          number_of_positional_args=1)

        # renew
        self.register_command_to_registry("renew-domain", RenewDomain,
                                          [ArgumentField("enum", RenewDomainEnumExtension, False)],
                                          number_of_positional_args=1)

        # extensions
        self.register_command_to_registry("credit-info-request", CreditInfoRequest)
        self.register_command_to_registry("send-auth-info-contact", SendAuthInfoContact,
                                          number_of_positional_args=1)
        self.register_command_to_registry("send-auth-info-domain", SendAuthInfoDomain,
                                          number_of_positional_args=1)
        self.register_command_to_registry("send-auth-info-keyset", SendAuthInfoKeyset,
                                          number_of_positional_args=1)
        self.register_command_to_registry("send-auth-info-nsset", SendAuthInfoNsset,
                                          number_of_positional_args=1)
        self.register_command_to_registry("test-nsset", TestNsset,
                                          number_of_positional_args=3)

        # delete
        self.register_command_to_registry("delete-contact", DeleteContact,
                                          number_of_positional_args=1)
        self.register_command_to_registry("delete-domain", DeleteDomain,
                                          number_of_positional_args=1)
        self.register_command_to_registry("delete-keyset", DeleteKeyset,
                                          number_of_positional_args=1)
        self.register_command_to_registry("delete-nsset", DeleteNsset,
                                          number_of_positional_args=1)

        # check
        self.register_command_to_registry("check-contact", CheckContact,
                                          number_of_positional_args=1)
        self.register_command_to_registry("check-domain", CheckDomain,
                                          number_of_positional_args=1)
        self.register_command_to_registry("check-keyset", CheckKeyset,
                                          number_of_positional_args=1)
        self.register_command_to_registry("check-nsset", CheckNsset,
                                          number_of_positional_args=1)

        # list
        self.register_command_to_registry("get-results", GetResults)
        self.register_command_to_registry("list-contacts", ListContacts,
                                          post_hooks=[get_results_hook])
        self.register_command_to_registry("list-domains", ListDomains,
                                          post_hooks=[get_results_hook])
        self.register_command_to_registry("list-domains-by-contact", ListDomainsByContact,
                                          post_hooks=[get_results_hook],
                                          number_of_positional_args=1)
        self.register_command_to_registry("list-domains-by-keyset", ListDomainsByKeyset,
                                          post_hooks=[get_results_hook],
                                          number_of_positional_args=1)
        self.register_command_to_registry("list-domains-by-nsset", ListDomainsByNsset,
                                          post_hooks=[get_results_hook],
                                          number_of_positional_args=1)
        self.register_command_to_registry("list-keysets", ListKeysets,
                                          post_hooks=[get_results_hook])
        self.register_command_to_registry("list-keysets-by-contact", ListKeysetsByContact,
                                          post_hooks=[get_results_hook],
                                          number_of_positional_args=1)
        self.register_command_to_registry("list-nssets", ListNssets,
                                          post_hooks=[get_results_hook])
        self.register_command_to_registry("list-nssets-by-contact", ListNssetsByContact,
                                          post_hooks=[get_results_hook],
                                          number_of_positional_args=1)
        self.register_command_to_registry("list-nssets-by-ns", ListNssetsByNs,
                                          post_hooks=[get_results_hook],
                                          number_of_positional_args=1)

        # help command
        help_command = HelpCommand(self.command_registry)
        self.command_registry.register_command(help_command)
