"""Command class responsible for recognising user input and arguments."""
import logging
from itertools import chain
from typing import TYPE_CHECKING, Any, Dict, Iterable, List, Optional, Type

from epplib.commands.base import Command as CommandClass

from eppic.models import ArgumentField, ExecutableCommand

from .argument_parser import ArgumentParser
from .argument_registry import ArgumentRegistry

if TYPE_CHECKING:
    from eppic.app import Application, CommandHook

LOGGER = logging.getLogger(__name__)


class Command:
    """Command class."""

    def __init__(
        self,
        name: str,
        command_class: Type[ExecutableCommand],
        fields: Iterable[ArgumentField],
        registry: ArgumentRegistry,
        extensions: Optional[Iterable[ArgumentField]] = None,
        pre_hooks: Optional[List["CommandHook"]] = None,
        post_hooks: Optional[List["CommandHook"]] = None,
        number_of_positional_args: int = 0,
    ) -> None:
        """Initialize the command from preprocessed parameters.

        Args:
            name: the name of the command.
            command_class: the class that is returned filled with arguments when the command is called.
            fields: an iterable of argument fields.
            registry: an ArgumentRegistry instance.
            extensions: an iterable of extensions fields.
            pre_hooks: a list of hooks to be run before a command.
            post_hooks: a list of hooks to be run after a command.
            number_of_positional_args: says that `n` first arguments are allowed to be positional. Defaults to 0.
        """
        extensions = extensions or ()
        self.name = name
        self._command_class = command_class
        self.arg_parser = ArgumentParser(chain(fields, extensions), registry, number_of_positional_args)
        self._extensions_list = tuple(f.name for f in extensions)
        self._pre_hooks = pre_hooks or []
        self._post_hooks = post_hooks or []

    def _pop_extensions(self, arguments: Dict[str, Any]) -> List[Any]:
        """Pop the extensions from arguments and return them.

        Args:
            arguments: arguments and extensions.
                This is an output argument: the extensions will be removed.

        Returns: the extracted extensions.

        """
        ret = []
        for key in list(arguments.keys()):
            if key in self._extensions_list:
                extension = arguments.pop(key)
                if extension is not None:
                    ret.append(extension)
        return ret

    def prepare_command(self, arguments: Dict[str, Any]) -> ExecutableCommand:
        """Build the Command Class and return it.

        Args:
            arguments: already parsed arguments of the command.

        """
        extensions = self._pop_extensions(arguments)
        command = self._command_class(**arguments)

        if isinstance(command, CommandClass):
            for extension in extensions:
                LOGGER.debug(f"Extension: {extension}")
                command.add_extension(extension)

        return command

    def pre_hooks(self, app: "Application", command: ExecutableCommand) -> None:
        """Run the pre-hooks."""
        for hook in self._pre_hooks:
            hook(app, command)

    def post_hooks(self, app: "Application", command: ExecutableCommand) -> None:
        """Run the post-hooks."""
        for hook in self._post_hooks:
            hook(app, command)
