"""Service Commands implementation."""
import logging
import sys
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import TYPE_CHECKING, List, Optional, cast

from eppic.models import SessionId
from eppic.settings import OutputFormat, Session, Settings

if TYPE_CHECKING:
    from eppic.app import Application

LOGGER = logging.getLogger(__name__)


class ServiceCommand(ABC):
    """Base class. Provides information to the app."""

    @abstractmethod
    def handle(self, app: "Application") -> None:
        """Handle the service command."""


class ServiceExit(ServiceCommand):
    """Allows the program to exit."""

    def handle(self, app: "Application") -> None:
        """Handle the service command."""
        app.epp_handler.logout()
        sys.exit(0)


@dataclass
class ServiceLogin(ServiceCommand):
    """Log into an existing session."""

    session_id: Optional[SessionId] = None

    def handle(self, app: "Application") -> None:
        """Handle the service command."""
        session = app.settings.get_session(self.session_id)

        if session is None:
            # `get_session` only returns `None` if `settigs.sessions` is an empty list.
            raise ValueError("No sessions exist.")
        else:
            app.epp_handler.session = session


class ServiceLogout(ServiceCommand):
    """Handle logout."""

    def handle(self, app: "Application") -> None:
        """Handle the service command."""
        app.epp_handler.logout()


@dataclass
class ServiceNewSession(ServiceCommand):
    """Create new Session instance."""

    base_session_id: Optional[SessionId] = None

    username: Optional[str] = None
    password: Optional[str] = None
    hostname: Optional[str] = None
    port: Optional[int] = None
    cert_file: Optional[str] = None
    key_file: Optional[str] = None
    obj_uris: Optional[List[str]] = None
    verify: Optional[bool] = None
    alias: Optional[str] = None

    def modify_session(self, session: Session) -> Session:
        """Modify the session from the object attributes.

        If an attribute is not set, take it from the base `session`.
        (`alias` is the exception of this -- if it isn't set, don't set it at all)
        (`obj_uris` is another exception -- it isn't compared to `None`, but to `bool`)
        """
        username = self.username if self.username is not None else session.username
        password = self.password if self.password is not None else session.password
        hostname = self.hostname if self.hostname is not None else session.hostname
        port = self.port if self.port is not None else session.port
        cert_file = self.cert_file if self.cert_file is not None else session.cert_file
        key_file = self.key_file if self.key_file is not None else session.key_file
        # obj_uris are set based on whether the're an empty list (which would be an invalid state)
        obj_uris = self.obj_uris if self.obj_uris else session.obj_uris
        verify = self.verify if self.verify is not None else session.verify
        alias = self.alias

        return Session(
            username=username, password=password, hostname=hostname,
            port=port, cert_file=cert_file, key_file=key_file, obj_uris=obj_uris,
            verify=verify, alias=alias
        )

    def handle(self, app: "Application") -> None:
        """Handle the service command."""
        if self.base_session_id is None:
            session = Session()
        else:
            session = cast(Session, app.settings.get_session(self.base_session_id))

        session = self.modify_session(session)
        session_number = app.add_session(session)
        LOGGER.info(f"Session added as number {session_number}.")
        app.epp_handler.session = session  # this performs automatic login if possible


@dataclass
class ServiceChangeSettings(ServiceCommand):
    """Change settings on the fly."""

    output_format: Optional[OutputFormat] = None
    autocomplete: Optional[bool] = None
    syntax_highlighting: Optional[bool] = None
    command_confirmation: Optional[bool] = None

    def edit_settings(self, settings: Settings) -> None:
        """Edit the settings with gathered arguments."""
        settings.output_format = self.output_format if self.output_format is not None else settings.output_format
        settings.interactive_input.autocomplete = self.autocomplete if self.autocomplete is not None \
            else settings.interactive_input.autocomplete
        settings.syntax_highlighting = self.syntax_highlighting if self.syntax_highlighting is not None \
            else settings.syntax_highlighting
        settings.command_confirmation = self.command_confirmation if self.command_confirmation is not None \
            else settings.command_confirmation

    def handle(self, app: "Application") -> None:
        """Handle the service command."""
        settings = app.settings
        self.edit_settings(settings)
        app.set_settings(settings)
