"""EPP Handler, carries the EPP session."""
import logging
from typing import Optional, cast

from epplib.client import Client
from epplib.commands import Login, Logout
from epplib.commands.base import Request
from epplib.exceptions import TransportError
from epplib.policy import Policy
from epplib.responses import Response, Result
from epplib.transport import SocketTransport

from .settings import Session

LOGGER = logging.getLogger(__name__)


class EppHandler:
    """Wrapper around the EPPlib."""

    def __init__(self, session: Optional[Session]) -> None:
        self._client: Optional[Client] = None
        self._session: Optional[Session] = None

        if session is not None:
            self.session = session

    @property
    def client(self) -> Client:
        """Return a client instance."""
        if self._client is None:
            if self.session.hostname is None:
                raise ValueError("Hostname not specified.")
            transport = SocketTransport(
                self.session.hostname, self.session.port,
                cert_file=self.session.cert_file,
                key_file=self.session.key_file,
                verify=self.session.verify
            )
            self._client = Client(transport)

        return self._client

    @property
    def logged_in(self) -> bool:
        """Is there a logged-in connection to the server."""
        return self._client is not None

    @property
    def session(self) -> Session:
        """Get session, if set."""
        if self._session is None:
            raise ValueError("Session not set.")

        return self._session

    @session.setter
    def session(self, session: Session) -> None:
        """Set a session."""
        self.logout()

        self._session = session

        # allows defining sessions without username and password
        # so user can log in manually later
        if session.username and session.password:
            self.login()

    def logout(self) -> None:
        """Send a logout command."""
        if not self.logged_in:
            return

        try:
            self._send_command(Logout())
        except Exception:
            LOGGER.warning("Logout failed.")

        try:
            self.client.close()
        except Exception:
            LOGGER.warning("Connection failed on close.")

        self._client = None

    def login(self) -> None:
        """Send a login command."""
        self.logout()

        username = cast(str, self.session.username)
        password = cast(str, self.session.password)

        try:
            self.client.connect()
            response = cast(Result, self.send_command(Login(cl_id=username, password=password,
                                                            obj_uris=self.session.obj_uris)))
            if response.code >= 2000:
                if response.code in (2200, 2501):
                    raise ValueError("Invalid username or password.")
                else:
                    raise ValueError("Login was unsuccessful.")
        except Exception:
            self.logout()
            raise

    @property
    def policy(self) -> Optional[Policy]:
        """Get `Policy` of the client. `None` if not logged in."""
        return self.client.policy if self.logged_in else None

    def send_command(self, command: Request) -> Response:
        """Send an EPP command to the server."""
        try:
            return self._send_command(command)
        except TransportError:
            LOGGER.error("Communication with the server failed. Logging out and closing connection...")
            self.logout()
            raise

    def _send_command(self, command: Request) -> Response:
        """Send an EPP command to the server."""
        LOGGER.info(f"Send command: {command}")
        response = self.client.send(command)

        return response
