"""Handle printing of output."""
import datetime
import json
import sys
from dataclasses import asdict, is_dataclass
from decimal import Decimal
from enum import Enum
from pprint import PrettyPrinter
from typing import Any, List, Optional, Sequence

import yaml
from epplib.responses.base import Response, Result
from epplib.responses.info import InfoContactResultData
from prompt_toolkit import print_formatted_text
from prompt_toolkit.formatted_text import PygmentsTokens
from pygments import lex
from pygments.lexers import YamlLexer

INDENT_MULTIPLIER = 4


class EppicSafeDumper(yaml.SafeDumper):
    """Custom YAML safe dumper with eppic specifics."""

    def represent_enum(self, member: Enum) -> Any:
        """Represent enum members as their values."""
        return self.represent_data(member.value)

    def represent_decimal(self, data: Decimal) -> Any:
        """Represent decimal as a string."""
        return self.represent_data(str(data))


EppicSafeDumper.add_representer(Decimal, EppicSafeDumper.represent_decimal)
EppicSafeDumper.add_multi_representer(str, EppicSafeDumper.represent_str)
EppicSafeDumper.add_multi_representer(Enum, EppicSafeDumper.represent_enum)


class EppicTextSafeDumper(EppicSafeDumper):
    """Custom YAML safe dumper for the text output."""

    def represent_decimal(self, data: Decimal) -> Any:
        """Represent decimal as a string."""
        return self.represent_scalar('tag:yaml.org,2002:float', str(data))

    def represent_binary_data(self, data: bytes) -> Any:
        """Represent binary data as a string."""
        try:
            return self.represent_data(str(data, encoding="utf-8"))
        except UnicodeDecodeError:
            return self.represent_data(str(data))


EppicTextSafeDumper.add_representer(Decimal, EppicTextSafeDumper.represent_decimal)
EppicTextSafeDumper.add_representer(bytes, EppicTextSafeDumper.represent_binary_data)


class OutputHandler:
    """Prints output."""

    def __init__(self, syntax_highlighting: bool = True) -> None:
        """Create a PP instance."""
        self.pp = PrettyPrinter(indent=INDENT_MULTIPLIER)
        self._syntax_highlighting = syntax_highlighting

    def _serialize(self, data: Any) -> Any:
        """Serialize res_data objects to simple types."""
        if isinstance(data, InfoContactResultData) and data.disclose is not None:
            data_dict = asdict(data)
            data_dict["disclose"] = {"disclosed": data.disclose.disclosed_fields,
                                     "undisclosed": data.disclose.hidden_fields}
            return data_dict
        if is_dataclass(data):
            return asdict(data)
        return data

    def _output(self, msg: Optional[str], code: Optional[int], data_list: List[Any]) -> None:
        if code is not None:
            print(f"{msg} ({code})\n")

        data_out = data_list[0] if len(data_list) == 1 and isinstance(data_list[0], dict) else {"data": data_list}
        data = yaml.dump(data_out, Dumper=EppicTextSafeDumper, allow_unicode=True)

        if self._syntax_highlighting:
            tokens = list(lex(data, lexer=YamlLexer()))
            print_formatted_text(PygmentsTokens(tokens))
        else:
            print(data)

    def output(self, res: Response) -> None:
        """Output the Result in res."""
        if isinstance(res, Result):
            res_data = res.res_data if res.res_data is not None else []
            msg = res.msg
            code = res.code
        else:
            res_data = [res]
            msg, code = None, None

        self._output(msg, code, [self._serialize(r) for r in res_data])


class JsonOutputHandler(OutputHandler):
    """Outputs JSON. Multiple responses are split by newlines."""

    def _output(self, msg: Optional[str], code: Optional[int], data_list: List[Any]) -> None:
        res = {"msg": msg, "code": code, "data": data_list}
        json.dump(res, sys.stdout, sort_keys=False, default=self._json_converter, ensure_ascii=False)
        print("")

    def _json_converter(self, obj: Any) -> Any:  # pragma: no branch
        """Convert complex types to JSON format."""
        if isinstance(obj, (datetime.datetime, datetime.date)):
            obj = obj.isoformat()
        if isinstance(obj, Decimal):
            obj = str(obj)
        if isinstance(obj, set):
            obj = sorted(obj)  # sort the set values to make the output predictible
        if isinstance(obj, bytes):
            try:
                obj = str(obj, encoding="utf-8")  # encode the obj in utf-8
            except UnicodeDecodeError:
                obj = str(obj)

        # only happens when we encounter an unknown object.
        return obj  # pragma: no cover


class YamlOutputHandler(OutputHandler):
    """Outputs YAML. Multiple responses are split into multiple documents."""

    def _output(self, msg: Optional[str], code: Optional[int], data_list: Sequence[Any]) -> None:
        res = {"msg": msg, "code": code, "data": data_list}
        yaml.dump(res, sys.stdout, sort_keys=False, explicit_start=True, explicit_end=True, allow_unicode=True,
                  Dumper=EppicSafeDumper)
