"""Handle Settings from several sources."""
from __future__ import annotations

import sys
from enum import Enum, unique
from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, TextIO, Tuple, Union, cast

import yaml
from epplib.constants import NAMESPACE
from pydantic import BaseModel, BaseSettings, Extra, Field, FilePath
from pydantic.env_settings import SettingsSourceCallable

from .input.input_handler import (CliArgsInputHandler, FileInputHandler, InputHandler, JsonInputHandler,
                                  PromptToolkitInputHandler, YamlInputHandler)
from .input.input_processor import DictInputProcessor, InputProcessor, StringInputProcessor
from .input.prompt_toolkit_handlers import Completer as CustomCompleter, Validator as CustomValidator
from .models import SessionId
from .output import JsonOutputHandler, OutputHandler, YamlOutputHandler

if TYPE_CHECKING:
    from .commands.command_registry import CommandRegistry

SETTINGS_FILE_LOCATIONS = ["~/.eppic.conf", "/etc/eppic/eppic.conf"]


def _parse_settings_files(settings: Settings) -> Dict[str, Any]:
    """Parse a yaml config file. Either get it from the settings or choose it from `SETTINGS_FILE_LOCATIONS`.

    Args:
        settings: Base settings object.
        file_path: File path to the config file.

    Returns: A dictionary with the parsed values.

    """
    settings_file = settings._settings_file

    if settings_file is None:
        for file_location in SETTINGS_FILE_LOCATIONS:
            file_path = Path(file_location).expanduser()
            if file_path.is_file():
                settings_file = file_path
                break
    else:
        settings_file = Path(settings_file).expanduser()

    if settings_file is None:
        return {}

    with open(settings_file, "r") as f:
        return cast(Dict[str, Any], yaml.safe_load(f))


@unique
class OutputFormat(str, Enum):
    """Defines types of input."""

    DEFAULT = "DEFAULT"
    JSON = "JSON"
    YAML = "YAML"

    def get_output_handler(self, syntax_highlighting: bool = True) -> OutputHandler:
        """Return an output handler."""
        if self == OutputFormat.JSON:
            return JsonOutputHandler()
        elif self == OutputFormat.YAML:
            return YamlOutputHandler()
        else:
            return OutputHandler(syntax_highlighting=syntax_highlighting)


@unique
class InputFormat(str, Enum):
    """Defines types of input."""

    DEFAULT = "DEFAULT"
    JSON = "JSON"
    YAML = "YAML"

    def get_input_handler(
        self, command: Optional[str], file_path: Optional[Path],
        command_registry: "CommandRegistry", history_path: Optional[Path],
        prompt: Union[str, Callable[[], str]] = "> ", syntax_highlighting: bool = True,
        autocomplete: bool = True
    ) -> InputHandler:
        """Generate an Input Handler."""
        if command is not None:
            return CliArgsInputHandler(command)
        elif file_path is None:
            input_processor = self.get_input_processor()
            completer = CustomCompleter(command_registry=command_registry, input_processor=input_processor)
            validator = CustomValidator(command_registry=command_registry, input_processor=input_processor)
            return PromptToolkitInputHandler(completer, validator, history_path=history_path, prompt=prompt,
                                             syntax_highlighting=syntax_highlighting, autocomplete=autocomplete)

        with open(file_path, "r") as f:
            return self.get_input_handler_based_on_format(f)

    def get_input_handler_based_on_format(self, file: TextIO) -> InputHandler:
        """Get one of the non-tty input handlers, only based on the format stored in `self`."""
        if self == InputFormat.JSON:
            return JsonInputHandler(file)
        if self == InputFormat.YAML:
            return YamlInputHandler(file)
        else:
            return FileInputHandler(file)

    def get_input_processor(self) -> InputProcessor:
        """Return an Input Processor."""
        if self in [InputFormat.JSON, InputFormat.YAML]:
            return DictInputProcessor()

        return StringInputProcessor()


class Session(BaseModel):
    """Session configuration info."""

    hostname: Optional[str] = None
    port: int = 700
    cert_file: Optional[FilePath] = None
    key_file: Optional[FilePath] = None
    obj_uris: List[str] = [NAMESPACE.NIC_CONTACT, NAMESPACE.NIC_DOMAIN, NAMESPACE.NIC_NSSET, NAMESPACE.NIC_KEYSET]
    username: Optional[str] = None
    password: Optional[str] = None
    verify: bool = True
    alias: Optional[str] = None

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Init model and get alias."""
        super().__init__(*args, **kwargs)

        if self.alias is None:
            self.alias = self.username


class InteractiveInputSettings(BaseModel):
    """Settings specific for interactive input."""

    autocomplete: bool = Field(default_factory=sys.stdin.isatty)
    history_file_path: Path = Path("~/.cache/eppic/autosuggest_history")


class Settings(BaseSettings):
    """Default Settings class."""

    _settings_file: Optional[Path] = None
    _session_id: Optional[SessionId] = None

    sessions: List[Session] = cast(List[Session], Field(default_factory=list))
    input_file: Optional[FilePath] = None
    input_format: InputFormat = InputFormat.DEFAULT
    output_format: OutputFormat = OutputFormat.DEFAULT
    logging: Optional[Dict] = None
    interactive_input: InteractiveInputSettings = InteractiveInputSettings()
    syntax_highlighting: bool = Field(default_factory=sys.stdin.isatty)

    command_confirmation: bool = False
    print_list_results_automatically: bool = True

    def __init__(
        self,
        *args: Any,
        settings_file: Optional[FilePath] = None,
        **kwargs: Any
    ):

        self._settings_file = Path(settings_file) if settings_file is not None else None
        super().__init__(*args, **kwargs)
        self._settings_file = Path(settings_file) if settings_file is not None else None

    def get_session(self, session_id: Optional[SessionId]) -> Optional[Session]:
        """Choose the session object settings."""
        if len(self.sessions) == 0 and session_id is not None:  # trying to choose a session, but none are defined
            raise ValueError("No sessions specified.")

        elif len(self.sessions) == 0:
            return None

        if session_id is not None:
            # specified by index
            if isinstance(session_id, int):
                if session_id >= len(self.sessions) or 0 > session_id:
                    raise ValueError("Specified session does not exist!")
                return self.sessions[session_id]

            # specified by username
            elif isinstance(session_id, str):
                for session in self.sessions:
                    if session.alias == session_id:
                        return session
                raise ValueError("Specified session does not exist!")
            else:
                raise TypeError("Wrong type of session id")

        # not specified - choose first
        else:
            return self.sessions[0]

    class Config:
        """Configure the Settings parsing."""

        env_prefix = "EPPIC_"
        extra = Extra.ignore
        underscore_attrs_are_private = True

        @classmethod
        def customise_sources(
            cls,
            init_settings: SettingsSourceCallable,
            env_settings: SettingsSourceCallable,
            file_secret_settings: SettingsSourceCallable,
        ) -> Tuple[SettingsSourceCallable, ...]:
            """List of source functions."""
            return (
                init_settings,
                cast(SettingsSourceCallable, _parse_settings_files),
                env_settings,
                file_secret_settings,
            )
