from unittest import TestCase
from unittest.mock import Mock, patch

from click.testing import CliRunner

from eppic.settings import Session

from .utils import DummyEppCommand, DummyResponse, GetAppMixin


class TestApplication(GetAppMixin, TestCase):

    def test_init(self):
        app = self.get_app()
        self.assertGreater(len(app.argument_registry._types), 0)
        self.assertGreater(len(app.command_registry._commands), 0)
        self.assertEppHandlerCreated()

    def test_prompt_filled(self):
        self.assertEqual(self.get_app()._prompt(), "foo@asdf > ")

    def test_prompt_empty(self):
        self.assertEqual(self.get_app(session=None)._prompt(), "> ")

    def test_execute_command(self):
        command = DummyEppCommand("asdf")
        self.get_app().execute(command)
        self.epp_handler.send_command.assert_called_once_with(command)

    def test_run_command(self):
        self.get_app().run_command("foobar asdf")
        self.epp_handler.send_command.assert_any_call(DummyEppCommand("asdf"))

    def test_run(self):
        self.get_app(command="foobar asdf").run()
        self.epp_handler.send_command.assert_any_call(DummyEppCommand("asdf"))
        self.assertEppHandlerExited()

    def test_run_error_logging(self):
        app = self.get_app(command="foobar asdf")
        self.epp_handler.send_command.side_effect = ValueError()
        with patch("eppic.app.LOGGER") as mock:
            app.run()
            mock.error.assert_called_once()
            self.assertEppHandlerExited()

    def test_run_parsing_fail(self):
        self.get_app(command="foobar asdf asdf").run()
        self.epp_handler.send_command.assert_not_called()
        self.assertEppHandlerExited()

    def test_pre_hooks(self):
        self.epp_handler.send_command.return_value = DummyResponse()
        app = self.get_app(foobar_pre_hooks=[m := Mock()])
        app.run_command("foobar asdf")
        m.assert_called_once()
        self.assertEqual(m.call_args.args[0], app)

    def test_post_hooks(self):
        self.epp_handler.send_command.return_value = DummyResponse()
        app = self.get_app(foobar_post_hooks=[m := Mock()])
        app.run_command("foobar asdf")
        m.assert_called_once()
        self.assertEqual(m.call_args.args[0], app)

    def test_confirm_positive(self):
        with CliRunner().isolation(input="y"):
            self.get_app(confirm=True).run_command("foobar asdf")
            self.epp_handler.send_command.assert_called_once_with(DummyEppCommand("asdf"))

    def test_confirm_negative(self):
        with CliRunner().isolation(input="n"):
            self.get_app(confirm=True).run_command("foobar asdf")
            self.epp_handler.send_command.assert_not_called()

    def test_confirm_service_commands(self):
        self.get_app(confirm=True, foobar_is_service=True).run_command("foobar asdf")
        self.assertFoobarServiceExecuted()

    def test_add_session(self):
        app = self.get_app()
        session = Session(hostname="hello", username="world")
        sess_id = app.add_session(session)  # session id is an index of the session in the array
        self.assertEqual(len(app.settings.sessions), sess_id + 1)
        self.assertEqual(session, app.settings.get_session(sess_id))
