import logging
import tempfile
from pathlib import Path
from typing import Any, Dict
from unittest import TestCase
from unittest.mock import patch

import yaml
from click.testing import CliRunner

from eppic.commands.service import ServiceNewSession
from eppic.main import main


@patch("eppic.main.Application")
class TestMain(TestCase):

    def setUp(self) -> None:
        self.tmp_dir = tempfile.TemporaryDirectory()
        self.addCleanup(self.tmp_dir.cleanup)
        self.tmp_file = Path(self.tmp_dir.name) / "settings.yaml"
        with open(Path(self.tmp_dir.name) / "key", "w") as f:
            f.write("foo")

        with open(Path(self.tmp_dir.name) / "cert", "w") as f:
            f.write("foo")

        self.content: Dict[str, Any] = {
            "sessions": [
                {
                    "hostname": "asdf",
                    "port": 1234,
                    "cert_file": str(Path(self.tmp_dir.name) / "cert"),
                    "key_file": str(Path(self.tmp_dir.name) / "key"),
                    "obj_uris": ["asdf"],
                    "username": "asdf",
                    "password": "ghjkl"
                }
            ],
        }
        self.write_settings(self.content)

    def write_settings(self, content: Any) -> None:
        with open(self.tmp_file, "w") as f:
            yaml.safe_dump(content, f)

    def tearDown(self) -> None:
        self.tmp_dir.cleanup()

    def test_settings_error(self, session):
        (Path(self.tmp_dir.name) / "cert").rename(Path(self.tmp_dir.name) / "crt")
        result = CliRunner().invoke(main, f"--config {self.tmp_file}")
        self.assertEqual(result.exit_code, 1)

    def test_main_run(self, session):
        result = CliRunner().invoke(main, f"--config {self.tmp_file}")
        self.assertEqual(result.exit_code, 0, msg=result.exception)
        session.assert_called_once()

    def test_main_w_command(self, session):
        result = CliRunner().invoke(main, f"--config {self.tmp_file} -- asdf sadf asdf")
        self.assertEqual(result.exit_code, 0)
        self.assertEqual(session.call_args.args[-1], "asdf sadf asdf")

    def test_main_w_input(self, session):
        result = CliRunner().invoke(main, f"--config {self.tmp_file} --input-file {self.tmp_file}")
        self.assertEqual(result.exit_code, 0)

    def test_main_w_logging_set(self, session):
        logging.basicConfig(level="INFO")
        self.content["logging"] = {"version": 1, "root": {"level": "DEBUG"}}
        self.write_settings(self.content)
        result = CliRunner().invoke(main, f"--config {self.tmp_file}")
        self.assertEqual(logging.getLogger().level, logging.DEBUG)
        self.assertEqual(result.exit_code, 0)

    def test_main_new_session(self, session):
        result = CliRunner().invoke(
            main, f"--config {self.tmp_file} --session-id foobar_ignored --hostname hello --port 123 --username foobar\
                    --password passwd --obj-uris asdf --obj-uris ghjkl -- foobar")
        self.assertEqual(result.exit_code, 0)
        session.return_value.execute.assert_called_once_with(
            ServiceNewSession(  # the `base_session_id` is unspecified
                username="foobar", password="passwd", hostname="hello", port=123,
                obj_uris=["asdf", "ghjkl"], alias=None)
        )
        self.assertEqual(session.call_args.kwargs["session"], None)

    def test_main_not_new_session(self, session):
        result = CliRunner().invoke(
            main, f"--config {self.tmp_file} -- foobar")
        self.assertEqual(result.exit_code, 0)
        session.return_value.execute.assert_not_called()
