import tempfile
from pathlib import Path
from typing import Any, Dict, cast
from unittest import TestCase
from unittest.mock import Mock, patch

from pydantic import ValidationError

from eppic.input.input_handler import (CliArgsInputHandler, FileInputHandler, JsonInputHandler,
                                       PromptToolkitInputHandler, YamlInputHandler)
from eppic.input.input_processor import DictInputProcessor, StringInputProcessor
from eppic.output import JsonOutputHandler, OutputHandler, YamlOutputHandler
from eppic.settings import SETTINGS_FILE_LOCATIONS, InputFormat, OutputFormat, Session, Settings, _parse_settings_files

SESSION = {  # nosec
    "hostname": "asdf",
    "port": 1234,
    "cert_file": "cert",
    "key_file": "key",
    "username": "asdf",
    "password": "ghjkl"
}


class YamlParserTest(TestCase):

    def setUp(self) -> None:
        self.homeconfig = tempfile.NamedTemporaryFile().name
        self.etcconfig = tempfile.NamedTemporaryFile().name
        self.customconfig = tempfile.NamedTemporaryFile().name
        with open(self.homeconfig, "w") as f:
            f.write("HOME")
        with open(self.etcconfig, "w") as f:
            f.write("ETC")
        with open(self.customconfig, "w") as f:
            f.write("CUSTOM")
        self.settings_mock = Mock()
        self.settings_mock._settings_file = self.customconfig
        self._settings_original = SETTINGS_FILE_LOCATIONS.copy()
        SETTINGS_FILE_LOCATIONS[0] = self.homeconfig
        SETTINGS_FILE_LOCATIONS[1] = self.etcconfig

    def tearDown(self):
        SETTINGS_FILE_LOCATIONS[0] = self._settings_original[0]
        SETTINGS_FILE_LOCATIONS[1] = self._settings_original[1]

    def test_w_custom_file(self) -> None:
        self.assertEqual(_parse_settings_files(self.settings_mock), "CUSTOM")

    def test_w_both_settings_files(self) -> None:
        self.settings_mock._settings_file = None
        self.assertEqual(_parse_settings_files(self.settings_mock), "HOME")

    def test_wo_home_settings_file(self) -> None:
        self.settings_mock._settings_file = None
        SETTINGS_FILE_LOCATIONS[0] += "/asdfasdf"
        self.assertEqual(_parse_settings_files(self.settings_mock), "ETC")

    def test_wo_settings_files(self) -> None:
        self.settings_mock._settings_file = None
        SETTINGS_FILE_LOCATIONS[0] += "/asdfasdf"
        SETTINGS_FILE_LOCATIONS[1] += "/asdfasdf"
        self.assertEqual(_parse_settings_files(self.settings_mock), {})


@patch("eppic.settings._parse_settings_files", new=lambda *args: dict())
class SettingsTest(TestCase):

    @patch("eppic.settings.Path.exists", new=lambda *args: True)
    def setUp(self) -> None:
        self.default_session = SESSION.copy()
        self.default_session["cert_file"] = self.create_key()
        self.default_session["key_file"] = self.create_key()
        self.default_session.pop("username")
        self.default_settings: Dict[str, Any] = {
            "sessions": [
                Session(**self.default_session, username="baz"),
                Session(**self.default_session, username="foo"),
                Session(**self.default_session, username="bar"),
            ]
        }

    def create_key(self):
        filename = tempfile.NamedTemporaryFile().name
        with open(filename, "w") as f:
            f.write("foobar")
        return filename

    def settings(self) -> Settings:
        return Settings(**self.default_settings)

    def test_no_sessions_at_all(self) -> None:
        self.default_settings["sessions"] = []
        settings = self.settings()
        self.assertIsNone(settings.get_session(None))

    def test_session_none(self) -> None:
        self.assertEqual(cast(Session, self.settings().get_session(None)).username, "baz")

    def test_session_but_no_specified(self):
        self.default_settings.pop("sessions")
        settings = self.settings()
        with self.assertRaisesRegex(ValueError, "No sessions specified\\."):
            settings.get_session("hello")

    def test_session_by_name(self) -> None:
        settings = self.settings()
        self.assertEqual(cast(Session, settings.get_session("foo")).username, "foo")

    def test_session_by_alias(self) -> None:
        self.default_settings["sessions"].append(Session(
            **self.default_session, username="notalias", alias="hello"
        ))
        self.assertEqual(self.settings().get_session("hello"), self.default_settings["sessions"][-1])
        self.assertRaises(ValueError, self.settings().get_session, "notalias")

    def test_session_wrong_type(self) -> None:
        self.assertRaises(TypeError, self.settings().get_session, 0.01)

    def test_session_by_name_wrong(self) -> None:
        settings = self.settings()
        self.assertRaises(ValueError, settings.get_session, "nonexistent")

    def test_session_by_number(self) -> None:
        settings = self.settings()
        self.assertEqual(cast(Session, settings.get_session(1)).username, "foo")

    def test_session_by_number_out_of_range(self) -> None:
        for i in [-1, 123456789, -123456789]:
            with self.subTest(i=i):
                settings = self.settings()
                self.assertRaises(ValueError, settings.get_session, i)

    def test_extra_settings(self) -> None:
        self.default_settings["foobar"] = "foobar"
        self.settings()

    def test_settings_file_set(self):
        settings = Settings(**self.default_settings, settings_file=Path("/"))
        self.assertEqual(settings._settings_file, Path("/"))

    def test_input_file_set(self):
        filename = tempfile.NamedTemporaryFile().name
        with open(filename, "w") as f:
            f.write("foobar")

        settings = Settings(**self.default_settings, input_file=filename)
        self.assertEqual(settings.input_file, Path(filename))


class SessionTest(TestCase):

    def setUp(self) -> None:
        self.default_session = SESSION.copy()
        self.default_session["cert_file"] = self.create_key()
        self.default_session["key_file"] = self.create_key()

    def create_key(self):
        filename = tempfile.NamedTemporaryFile().name
        with open(filename, "w") as f:
            f.write("foobar")
        return filename

    def test_session_file_validators_false(self) -> None:
        tmpdir = tempfile.TemporaryDirectory()
        self.default_session["cert_file"] = Path(tmpdir.name)
        self.addCleanup(tmpdir.cleanup)
        self.assertRaises(ValidationError, Session, **self.default_session)

        self.default_session["cert_file"] /= "asdf"  # type: ignore
        self.assertRaises(ValidationError, Session, **self.default_session)

    def test_session_no_client(self) -> None:
        self.default_session.pop("username")
        self.default_session.pop("password")
        session = Session(**self.default_session)
        self.assertIsNone(session.username)
        self.assertIsNone(session.password)

    def test_session_independent_obj_uris(self):
        session1 = Session(**self.default_session)
        session2 = Session(**self.default_session)
        session2.obj_uris.pop(0)
        self.assertNotEqual(len(session1.obj_uris), len(session2.obj_uris))


class GetOutputHandlerTest(TestCase):
    def test_json(self):
        self.assertIsInstance(OutputFormat.JSON.get_output_handler(), JsonOutputHandler)

    def test_yaml(self):
        self.assertIsInstance(OutputFormat.YAML.get_output_handler(), YamlOutputHandler)

    def test_default(self):
        self.assertIsInstance(OutputFormat.DEFAULT.get_output_handler(), OutputHandler)


class GetInputHandlerTest(TestCase):

    def test_command(self):
        self.assertIsInstance(
            InputFormat.DEFAULT.get_input_handler("asdf", None, Mock(), Path("asdf")),
            CliArgsInputHandler
        )

    def test_prompt_toolkit(self):
        tmp = tempfile.TemporaryDirectory()
        self.addCleanup(tmp.cleanup)
        file = Path(tmp.name) / "file"
        with open(file, "w") as w:
            w.write("asdf")
        ih = InputFormat.DEFAULT.get_input_handler(None, None, Mock(), file)
        self.assertIsInstance(ih, PromptToolkitInputHandler)
        self.assertIsNotNone(cast(PromptToolkitInputHandler, ih).prompt.completer)

    def test_prompt_toolkit_no_completion(self):
        tmp = tempfile.TemporaryDirectory()
        self.addCleanup(tmp.cleanup)
        file = Path(tmp.name) / "file"
        with open(file, "w") as w:
            w.write("asdf")
        ih = InputFormat.DEFAULT.get_input_handler(None, None, Mock(), file, autocomplete=False)
        self.assertIsInstance(ih, PromptToolkitInputHandler)
        self.assertIsNone(cast(PromptToolkitInputHandler, ih).prompt.completer)

    def test_default(self):
        tmp = tempfile.TemporaryDirectory()
        self.addCleanup(tmp.cleanup)
        file = Path(tmp.name) / "file"
        with open(file, "w") as w:
            w.write("asdf")
        self.assertIsInstance(
            InputFormat.DEFAULT.get_input_handler(None, file, Mock(), Path("")),
            FileInputHandler
        )

    def test_json(self):
        tmp = tempfile.TemporaryDirectory()
        self.addCleanup(tmp.cleanup)
        file = Path(tmp.name) / "file"
        with open(file, "w") as w:
            w.write("{'asdf': 'asdf'}")
        self.assertIsInstance(
            InputFormat.JSON.get_input_handler(None, file, Mock(), Path("")),
            JsonInputHandler
        )

    def test_yaml(self):
        tmp = tempfile.TemporaryDirectory()
        self.addCleanup(tmp.cleanup)
        file = Path(tmp.name) / "file"
        with open(file, "w") as w:
            w.write("asdf: asdf")
        self.assertIsInstance(
            InputFormat.YAML.get_input_handler(None, file, Mock(), Path("")),
            YamlInputHandler
        )


class GetInputProcessorTest(TestCase):

    def test_dict(self):
        for dict_type in [InputFormat.JSON, InputFormat.YAML]:
            with self.subTest(type=dict_type):
                self.assertIsInstance(dict_type.get_input_processor(), DictInputProcessor)

    def test_string(self):
        for dict_type in [InputFormat.DEFAULT]:
            with self.subTest(type=dict_type):
                self.assertIsInstance(dict_type.get_input_processor(), StringInputProcessor)
