from dataclasses import dataclass
from typing import Any, List, Mapping, Optional, Type, cast
from unittest import TestCase
from unittest.mock import Mock, patch

from epplib.commands.base import Command as EppCommand
from epplib.responses.base import Response
from lxml.etree import Element

from eppic.app import Application
from eppic.commands.service import ServiceCommand
from eppic.settings import InputFormat, OutputFormat, Session, Settings

DEFAULT_SESSIONS = [
    Session(hostname="asdf", port=123, username="foo",
            password="passwd", obj_uris=["asdf"], cert_file=__file__, key_file=__file__),
    Session(hostname="asdf", port=123, username="asdff",
            password="asdf", obj_uris=["asdf"], cert_file=__file__, key_file=__file__),
]


@dataclass
class DummyServiceCommand(ServiceCommand):

    foobar: str
    baz: Optional[int] = None

    handle = Mock()


@dataclass
class DummyEppCommand(EppCommand):

    foobar: str
    baz: Optional[int] = None

    def xml(self, *args, **kwargs):
        """Return xml."""
        return Mock(spec=Element)

    def _get_command_payload(self) -> None:
        """Implement dummy method."""


@dataclass
class DummyResponse(Response):

    @classmethod
    def _extract_payload(cls, element: Any) -> Mapping[str, Any]:  # pragma: no cover
        """Implement dummy method."""
        return {}


class GetAppMixin:

    default_sessions = DEFAULT_SESSIONS

    def get_settings(self, *, confirm: bool = False, sessions: List[Session] = DEFAULT_SESSIONS,
                     syntax_highlighting: bool = True, autocomplete: bool = True,
                     input_format: InputFormat = InputFormat.DEFAULT, output_format: OutputFormat = OutputFormat.DEFAULT
                     ) -> Settings:
        """Get settings."""
        settings = Settings(sessions=sessions, output_format=output_format, command_confirmation=confirm,
                            input_format=input_format, syntax_highlighting=syntax_highlighting)
        settings.interactive_input.autocomplete = autocomplete
        return settings

    def get_app(self, *, session=DEFAULT_SESSIONS[0], command=None, foobar_pre_hooks=[], foobar_post_hooks=[],
                foobar_is_service=False, **kwargs):
        app = Application(self.get_settings(**kwargs), command, session)
        foobar_spec = DummyEppCommand if not foobar_is_service else DummyServiceCommand
        app.register_command_to_registry("foobar", cast(Type[DummyEppCommand], foobar_spec), None,
                                         pre_hooks=foobar_pre_hooks, post_hooks=foobar_post_hooks,
                                         number_of_positional_args=1)
        self.epp_handler.session = session
        self.epp_handler.logged_in = session is not None
        return app

    def setUp(self):
        # Patch `EppHandler` throughout the tests.
        patcher = patch("eppic.app.EppHandler", autospec=True)
        cast(TestCase, self).addCleanup(patcher.stop)
        self.epp_handler_cls = patcher.start()
        self.epp_handler = self.epp_handler_cls.return_value
        self.epp_handler.send_command.return_value = DummyResponse()

    def assertEppHandlerCreated(self):
        self.epp_handler_cls.assert_called()

    def assertEppHandlerExited(self) -> None:
        self.epp_handler.logout.assert_called()

    def assertFoobarServiceExecuted(self) -> None:
        DummyServiceCommand.handle.assert_called()
